import board, busio, time, usb_hid, json, neopixel
from adafruit_hid.keyboard import Keyboard
from adafruit_hid.keycode import Keycode
from adafruit_hid.mouse import Mouse
import adafruit_vl53l0x

config = {}
try:
    with open("config.json", "r") as config_file:
        config = json.load(config_file)
except (OSError, ValueError) as e:
    print(f"Config error: {e}. Using defaults.")
    config = {
        "proximity_threshold": 10, "triggered_key": "TAB",
        "key_press_duration_seconds": 0.1, "loop_delay_seconds": 0.25,
        "only_closer_trigger": True, "enable_debug_prints": False,
        "enable_neopixel_feedback": False, "neopixel_color_rgb": [0, 10, 0],
        "hold_key_until_proximity_change": False, "debounce_time_ms": 200,
        "key_action": {"type": "single_key", "keys": ["TAB"]},
        "mouse_action": {"type": "none", "scroll_amount": 1}
    }

keycode_map = {
    "A": Keycode.A, "B": Keycode.B, "C": Keycode.C, "D": Keycode.D, "E": Keycode.E,
    "F": Keycode.F, "G": Keycode.G, "H": Keycode.H, "I": Keycode.I, "J": Keycode.J,
    "K": Keycode.K, "L": Keycode.L, "M": Keycode.M, "N": Keycode.N, "O": Keycode.O,
    "P": Keycode.P, "Q": Keycode.Q, "R": Keycode.R, "S": Keycode.S, "T": Keycode.T,
    "U": Keycode.U, "V": Keycode.V, "W": Keycode.W, "X": Keycode.X, "Y": Keycode.Y,
    "Z": Keycode.Z, "ONE": Keycode.ONE, "TWO": Keycode.TWO, "THREE": Keycode.THREE,
    "FOUR": Keycode.FOUR, "FIVE": Keycode.FIVE, "SIX": Keycode.SIX, "SEVEN": Keycode.SEVEN,
    "EIGHT": Keycode.EIGHT, "NINE": Keycode.NINE, "ZERO": Keycode.ZERO,
    "F1": Keycode.F1, "F2": Keycode.F2, "F3": Keycode.F3, "F4": Keycode.F4,
    "F5": Keycode.F5, "F6": Keycode.F6, "F7": Keycode.F7, "F8": Keycode.F8,
    "F9": Keycode.F9, "F10": Keycode.F10, "F11": Keycode.F11, "F12": Keycode.F12,
    "ENTER": Keycode.ENTER, "RETURN": Keycode.RETURN, "ESCAPE": Keycode.ESCAPE,
    "BACKSPACE": Keycode.BACKSPACE, "TAB": Keycode.TAB, "SPACE": Keycode.SPACE,
    "MINUS": Keycode.MINUS, "EQUALS": Keycode.EQUALS, "LEFT_BRACKET": Keycode.LEFT_BRACKET,
    "RIGHT_BRACKET": Keycode.RIGHT_BRACKET, "BACKSLASH": Keycode.BACKSLASH,
    "SEMICOLON": Keycode.SEMICOLON, "QUOTE": Keycode.QUOTE, "GRAVE_ACCENT": Keycode.GRAVE_ACCENT,
    "COMMA": Keycode.COMMA, "PERIOD": Keycode.PERIOD, "FORWARD_SLASH": Keycode.FORWARD_SLASH,
    "CAPS_LOCK": Keycode.CAPS_LOCK, "SCROLL_LOCK": Keycode.SCROLL_LOCK,
    "PRINT_SCREEN": Keycode.PRINT_SCREEN, "PAUSE": Keycode.PAUSE, "INSERT": Keycode.INSERT,
    "HOME": Keycode.HOME, "PAGE_UP": Keycode.PAGE_UP, "DELETE": Keycode.DELETE,
    "END": Keycode.END, "PAGE_DOWN": Keycode.PAGE_DOWN, "RIGHT_ARROW": Keycode.RIGHT_ARROW,
    "LEFT_ARROW": Keycode.LEFT_ARROW, "DOWN_ARROW": Keycode.DOWN_ARROW, "UP_ARROW": Keycode.UP_ARROW,
    "KEYPAD_FORWARD_SLASH": Keycode.KEYPAD_FORWARD_SLASH, "KEYPAD_ASTERISK": Keycode.KEYPAD_ASTERISK,
    "KEYPAD_MINUS": Keycode.KEYPAD_MINUS, "KEYPAD_PLUS": Keycode.KEYPAD_PLUS,
    "KEYPAD_ENTER": Keycode.KEYPAD_ENTER, "KEYPAD_ONE": Keycode.KEYPAD_ONE,
    "KEYPAD_TWO": Keycode.KEYPAD_TWO, "KEYPAD_THREE": Keycode.KEYPAD_THREE,
    "KEYPAD_FOUR": Keycode.KEYPAD_FOUR, "KEYPAD_FIVE": Keycode.KEYPAD_FIVE,
    "KEYPAD_SIX": Keycode.KEYPAD_SIX, "KEYPAD_SEVEN": Keycode.KEYPAD_SEVEN,
    "KEYPAD_EIGHT": Keycode.KEYPAD_EIGHT, "KEYPAD_NINE": Keycode.KEYPAD_NINE,
    "KEYPAD_ZERO": Keycode.KEYPAD_ZERO, "KEYPAD_PERIOD": Keycode.KEYPAD_PERIOD,
    "LEFT_CONTROL": Keycode.LEFT_CONTROL, "LEFT_SHIFT": Keycode.LEFT_SHIFT,
    "LEFT_ALT": Keycode.LEFT_ALT, "LEFT_GUI": Keycode.LEFT_GUI,
    "RIGHT_CONTROL": Keycode.RIGHT_CONTROL, "RIGHT_SHIFT": Keycode.RIGHT_SHIFT,
    "RIGHT_ALT": Keycode.RIGHT_ALT, "RIGHT_GUI": Keycode.RIGHT_GUI,
}

PROXIMITY_THRESHOLD = config["proximity_threshold"]
KEY_PRESS_DURATION = config["key_press_duration_seconds"]
LOOP_DELAY = config["loop_delay_seconds"]
ONLY_CLOSER_TRIGGER = config["only_closer_trigger"]
ENABLE_DEBUG_PRINTS = config["enable_debug_prints"]
ENABLE_NEOPIXEL_FEEDBACK = config["enable_neopixel_feedback"]
NEOPIXEL_COLOR = tuple(config["neopixel_color_rgb"])
HOLD_KEY_UNTIL_PROXIMITY_CHANGE = config["hold_key_until_proximity_change"]
DEBOUNCE_TIME_MS = config["debounce_time_ms"]

KEY_ACTION_TYPE = config["key_action"].get("type", "single_key")
RAW_KEYS_LIST = config["key_action"].get("keys", ["TAB"])

ACTION_KEYCODES = []
for key_str in RAW_KEYS_LIST:
    keycode = keycode_map.get(key_str.upper())
    if keycode: ACTION_KEYCODES.append(keycode)
    else: print(f"Warning: Unknown key '{key_str}'. Using TAB."); ACTION_KEYCODES.append(Keycode.TAB)

if not ACTION_KEYCODES and (KEY_ACTION_TYPE == "single_key" or KEY_ACTION_TYPE == "combination"):
    print("Error: No valid keys for single/combination. Defaulting to TAB.")
    ACTION_KEYCODES = [Keycode.TAB]
elif not ACTION_KEYCODES and KEY_ACTION_TYPE == "sequence":
    print("Error: No valid keys for sequence. Defaulting to 'HELLO'.")
    ACTION_KEYCODES = [Keycode.H, Keycode.E, Keycode.L, Keycode.L, Keycode.O]

MOUSE_ACTION_TYPE = config["mouse_action"].get("type", "none")
MOUSE_SCROLL_AMOUNT = config["mouse_action"].get("scroll_amount", 1)

print(f"Config loaded: Threshold={PROXIMITY_THRESHOLD}, Key Action={KEY_ACTION_TYPE}, Mouse Action={MOUSE_ACTION_TYPE}, Hold={HOLD_KEY_UNTIL_PROXIMITY_CHANGE}, Debounce={DEBOUNCE_TIME_MS}ms, Sensor: VL53L0X, NeoPixel Enabled: {ENABLE_NEOPIXEL_FEEDBACK}, NeoPixel Color: {NEOPIXEL_COLOR}")

try:
    i2c = busio.I2C(board.SCL1, board.SDA1)
except ValueError as e:
    print(f"I2C error: {e}. Check STEMMA QT.")
    while True: time.sleep(1)

vl53 = adafruit_vl53l0x.VL53L0X(i2c)
keyboard = Keyboard(usb_hid.devices)
mouse = Mouse(usb_hid.devices)

pixel = neopixel.NeoPixel(board.NEOPIXEL, 1, brightness=0.5, auto_write=False)

if not ENABLE_NEOPIXEL_FEEDBACK:
    pixel.brightness = 0.0
    pixel.fill((0,0,0))
    pixel.show()
else:
    pixel.brightness = 0.5
    pixel.fill((0,0,0))
    pixel.show()

key_is_held = False
proximity_at_press = 0
last_range = vl53.range
last_trigger_time = 0

while True:
    try:
        current_range = vl53.range
        if current_range > 8000:
            if ENABLE_DEBUG_PRINTS: print("Out of range.")
            time.sleep(LOOP_DELAY)
            continue

        if ENABLE_DEBUG_PRINTS: print(f"Current range: {current_range}mm")

        change_detected = False
        if ONLY_CLOSER_TRIGGER:
            if (last_range - current_range) >= PROXIMITY_THRESHOLD:
                change_detected = True
        else:
            if abs(current_range - last_range) >= PROXIMITY_THRESHOLD:
                change_detected = True

        current_time_ms = time.monotonic_ns() // 1_000_000

        if HOLD_KEY_UNTIL_PROXIMITY_CHANGE:
            if not key_is_held and change_detected and (current_time_ms - last_trigger_time) > DEBOUNCE_TIME_MS:
                if MOUSE_ACTION_TYPE != "none":
                    if MOUSE_ACTION_TYPE == "scroll_up":
                        mouse.move(wheel=MOUSE_SCROLL_AMOUNT)
                        print(f"Scrolled up by {MOUSE_SCROLL_AMOUNT} units and held!")
                    elif MOUSE_ACTION_TYPE == "scroll_down":
                        mouse.move(wheel=-MOUSE_SCROLL_AMOUNT)
                        print(f"Scrolled down by {MOUSE_SCROLL_AMOUNT} units and held!")
                    elif MOUSE_ACTION_TYPE == "left_click":
                        mouse.press(1)
                        print("Left mouse button pressed and held!")
                    elif MOUSE_ACTION_TYPE == "right_click":
                        mouse.press(2)
                        print("Right mouse button pressed and held!")
                    if ENABLE_NEOPIXEL_FEEDBACK:
                        pixel.brightness = 0.5
                        pixel.fill(NEOPIXEL_COLOR)
                        pixel.show()
                else:
                    if KEY_ACTION_TYPE == "single_key" or KEY_ACTION_TYPE == "combination":
                        for keycode in ACTION_KEYCODES: keyboard.press(keycode)
                        print(f"Keys {RAW_KEYS_LIST} pressed and held!")
                    elif KEY_ACTION_TYPE == "sequence":
                        keyboard.press(ACTION_KEYCODES[0])
                        print(f"Sequence first key {RAW_KEYS_LIST[0]} pressed and held!")
                    if ENABLE_NEOPIXEL_FEEDBACK:
                        pixel.brightness = 0.5
                        pixel.fill(NEOPIXEL_COLOR)
                        pixel.show()

                key_is_held = True
                proximity_at_press = current_range
                last_trigger_time = current_time_ms

            elif key_is_held:
                release_condition_met = False
                if ONLY_CLOSER_TRIGGER:
                    if (current_range - proximity_at_press) >= PROXIMITY_THRESHOLD:
                        release_condition_met = True
                else:
                    if abs(current_range - proximity_at_press) >= PROXIMITY_THRESHOLD:
                        release_condition_met = True

                if release_condition_met:
                    if MOUSE_ACTION_TYPE != "none":
                        mouse.release_all()
                        print("Mouse action released!")
                    else:
                        keyboard.release_all()
                        print(f"Keys {RAW_KEYS_LIST} released!")

                    if ENABLE_NEOPIXEL_FEEDBACK:
                        pixel.fill((0, 0, 0))
                        pixel.show()
                    key_is_held = False
                    proximity_at_press = 0
        else: # Not HOLD_KEY_UNTIL_PROXIMITY_CHANGE
            if change_detected and (current_time_ms - last_trigger_time) > DEBOUNCE_TIME_MS:
                if MOUSE_ACTION_TYPE != "none":
                    if MOUSE_ACTION_TYPE == "scroll_up":
                        mouse.move(wheel=MOUSE_SCROLL_AMOUNT)
                        print(f"Scrolled up by {MOUSE_SCROLL_AMOUNT} units!")
                    elif MOUSE_ACTION_TYPE == "scroll_down":
                        mouse.move(wheel=-MOUSE_SCROLL_AMOUNT)
                        print(f"Scrolled down by {MOUSE_SCROLL_AMOUNT} units!")
                    elif MOUSE_ACTION_TYPE == "left_click":
                        mouse.click(1)
                        print("Left mouse click!")
                    elif MOUSE_ACTION_TYPE == "right_click":
                        mouse.click(2)
                        print("Right mouse click!")
                    # NeoPixel ON/OFF logic for STANDARD mouse action (instant flash)
                    if ENABLE_NEOPIXEL_FEEDBACK:
                        pixel.brightness = 0.5
                        pixel.fill(NEOPIXEL_COLOR)
                        pixel.show()
                        time.sleep(0.05) # Small flash duration
                        pixel.fill((0,0,0))
                        pixel.show()
                else:
                    if KEY_ACTION_TYPE == "single_key" or KEY_ACTION_TYPE == "combination":
                        for keycode in ACTION_KEYCODES: keyboard.press(keycode)
                        print(f"Keys {RAW_KEYS_LIST} pressed!")
                        if ENABLE_NEOPIXEL_FEEDBACK:
                            pixel.brightness = 0.5
                            pixel.fill(NEOPIXEL_COLOR)
                            pixel.show()
                        time.sleep(KEY_PRESS_DURATION)
                        keyboard.release_all()
                        print(f"Keys {RAW_KEYS_LIST} released!")
                        if ENABLE_NEOPIXEL_FEEDBACK:
                            pixel.fill((0, 0, 0))
                            pixel.show()
                    elif KEY_ACTION_TYPE == "sequence":
                        for keycode in ACTION_KEYCODES:
                            keyboard.press(keycode)
                            print(f"Key {keycode} pressed!")
                            if ENABLE_NEOPIXEL_FEEDBACK:
                                pixel.brightness = 0.5
                                pixel.fill(NEOPIXEL_COLOR)
                                pixel.show()
                            time.sleep(KEY_PRESS_DURATION)
                            keyboard.release(keycode)
                            print(f"Key {keycode} released!")
                            if ENABLE_NEOPIXEL_FEEDBACK:
                                pixel.fill((0, 0, 0))
                                pixel.show()
                            time.sleep(0.05)
                last_trigger_time = current_time_ms

        last_range = current_range
        time.sleep(LOOP_DELAY)

    except OSError as e:
        print(f"Sensor error: {e}")
        try:
            i2c = busio.I2C(board.SCL1, board.SDA1)
            vl53 = adafruit_vl53l0x.VL53L0X(i2c)
            print("Sensor re-initialized.")
        except ValueError:
            print("Failed to re-initialize sensor. Check connections.")
            time.sleep(1)